#include<stdio.h>
#include<stdlib.h>
#include<omp.h>


int conway_par(int N, int M, int (*current_grid)[M], int (*next_grid)[M]) {
  int total = 0;
  int tid = omp_get_thread_num();
  int nthr = omp_get_num_threads();
  int total_work = (N-2) * (M-2);
  int total_chunk = total_work / nthr;
  int i_chunk = (N-2 + nthr - 1) / nthr;
  int j_chunk = (M-2 + nthr - 1) / nthr;
  int total_start = tid * total_chunk;
  int total_end = total_start + total_chunk;
  if (tid == nthr - 1) total_end = total_work;
  int i_start = 1 + total_start / (M - 2);
  int i_end = 1 + total_end / (M - 2);
  int j_start = 1 + total_start % (M - 2);
  int j_end = 1 + total_end % (M - 2);
  //int i_start = 1 + i_chunk * tid;
  //int i_end = i_start + i_chunk;
  //if (i_end > N - 1) i_end = N - 1;
  //i_end -= 1;
  //int j_start = 1;
  //int j_end = M - 1;
  //#pragma omp for collapse(2)
  for (int i = i_start; i <= i_end; i++) {
    int first_j = i == i_start ? j_start : 1;
    int last_j = i == i_end ? j_end : M - 1;
    for (int j = first_j; j < last_j; j++) {
      int count = 0;
      for (int ii = -1; ii <= 1; ii++)
        for (int jj = -1; jj <= 1; jj++)
          if (ii != 0 || jj != 0)
            count += current_grid[i+ii][j+jj];
      if (count == 3)
        next_grid[i][j] = 1;
      else if (count == 2)
        next_grid[i][j] = current_grid[i][j];
      else
        next_grid[i][j] = 0;
      total += next_grid[i][j];
    }
  }
  return total;
}

void print_grid(int N, int M, int* grid) {
  for (int i = 0; i < N; i++) {
    for (int j = 0; j < M; j++) {
      if (grid[M*i+j]) printf("#");
      else printf(" ");
    }
    printf("\n");
  }
}

int main(int argc, char** argv) {
  if (argc != 5) {
    printf("Usage: %s <grid_size_x> <grid_size_y> <timesteps> <fill-factor>\n", argv[0]);
    return;
  }
  int N = atoi(argv[1]);
  int M = atoi(argv[2]);
  int T = atoi(argv[3]);
  double fill = atof(argv[4]);
  int *grid_0 = calloc(N, M * sizeof(int));
  int *grid_1 = calloc(N, M * sizeof(int));
  for (int i = 1; i < N-1; i++)
    for (int j = 1; j < M-1; j++)
      grid_0[M*i+j] = rand() < (RAND_MAX * fill);
  //print_grid(N, M, grid_0);
  //for (int t = 0; t < T; t++) {
  //  int count = 0;
  //  #pragma omp parallel reduction(+:count)
  //  count += conway_par(N, M, (void*) grid_0, (void*) grid_1);
  //  print_grid(N, M, grid_1);
  //  printf("%d %d\n", t, count);
  //  int *tmp = grid_0;
  //  grid_0 = grid_1;
  //  grid_1 = tmp;
  //}
  int count = 0;
  #pragma omp parallel
  for (int t = 0; t < T; t++) {
    #pragma omp atomic
    count += conway_par(N, M, (void*) grid_0, (void*) grid_1);
    #pragma omp barrier
    if (omp_get_thread_num() == 0) {
      //print_grid(N, M, grid_1);
      printf("%d %d\n", t, count);
      int *tmp = grid_0;
      grid_0 = grid_1;
      grid_1 = tmp;
      count = 0;
    }
    #pragma omp barrier
  }
  free(grid_0);
  free(grid_1);
}
